//===-------------- PhiAnalysis_Fer.cpp - Printing prototype --------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it groups virtual registers that are related by phi functions.
//
// Date: August 1th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class groups together virtual registers that are related by phi
// functions in the control flow graph. Such virtuals describe the same
// variable in the original program. A good register allocator should try to
// assign the same physical register to these virtuals. For instance, if
// v1 := PHI (v2, v3, v3), then {v1, v2, v3} are phi related virtuals.
//
//===----------------------------------------------------------------------===//

#ifndef PHI_ANALYSIS_H
#define PHI_ANALYSIS_H

#include "llvm/PassSupport.h"

namespace llvm {

  class DefUseSites_Fer;
  class MachineFunction;
  class MachineInstr;
  class RegMapping_Fer;
  class SeparatePhiLiveRanges_Fer;

  class PhiAnalysis_Fer : public MachineFunctionPass {

    public:

    /// reporst that this pass does not change anything on the control flow
    /// graph.
    virtual void getAnalysisUsage(AnalysisUsage & AU) const;


    /// Informs the name of the pass, so it can appear in tools like llc.
    virtual const char * getPassName() const {
        return "Phi Analysis - Fernando.";
    }


    /// This method returns a vector {c_0, c_1, ..., c_n-1}, where each cell
    /// is a color that best fits v_reg for register allocation purposes. This
    /// means that, if v_reg is assigned c_0, more phi-function entries will be
    /// coalesced than if v_reg is assigned c_1 and so on.
    const std::vector<unsigned> * get_preferences(
        unsigned v_reg,
        const RegMapping_Fer * rmap
    );


    /// This method returns true if v_reg is part of a phi function somewhere
    /// in the control flow graph, and false otherwise. 
    bool is_phi_related(unsigned v_reg);


    /// This method finds all the definition sites for phi-functions. Given a
    /// phi-function i, the set of its definition sites is formed by
    /// instructions that define the parameters of i. The algorithm is
    /// iterative, because there may be recursive definitions between phi
    /// functions.
    void find_definition_sites(DefUseSites_Fer *du);


    /// Prints the set of phi-related virtuals
    void dump();


    /// This method returns all the definition sites of the given virtual. An
    /// error will be produced if the virtual is not in the definition list of
    /// a phi-equation.
    std::set<unsigned> & get_def_sites(unsigned v_reg);


    /// This method returns true if the given virtual (or, for the same effect,
    /// any other phi-related virtual) has been mapped to a stack slot.
    bool has_stack_slot(unsigned v);


    /// This method returns the stack slot that has been assigned to virtual
    /// v, if it has been spilled to memory. All the phi-related virtuals have
    /// the same stack slot.
    int get_stack_slot(unsigned v);


    /// This method maps the virtual v to the stack slot s. It also has the
    /// effect of assigning s to all other virtuals that are phi-related to v.
    /// If v has already been mapped to a stack slot, the function will produce
    /// an error.
    void map_to_stack_slot(unsigned v, int s);


    /// This function merges the phi-group of v1 with the phi-group of v2, so
    /// that each virtual in any of these groups will be phi-related.
    void merge(unsigned v1, unsigned v2);


    /// This method returns the set of virtuals that are phi-related to v_reg
    std::set<unsigned> & get_phi_group(unsigned v_reg);


    /// This method returns all the phi-function in which the virtual v_reg is
    /// used.
    std::vector<MachineInstr *> & get_use_list(unsigned v_reg);

    private:

    /// This integer holds the index of the virtual with the biggest id. It
    /// is useful when scanning the vector rltd_list.
    unsigned max_virtual;


    /// Number of machine registers in the target architecture. It will never
    /// change during program execution.
    unsigned num_p_regs;


    /// This vector contains a collection of sets. Each set groups together all
    /// the virtuals that are related by some phi-function. In theory, two
    /// virtuals are phi-related if they were part of the same variable
    /// before the SSA-deconstruction phase. This information is lost in LLVM,
    /// and we do an approaximation here.
    std::vector< std::set<unsigned> * > phi_related_virtuals;


    /// Each cell of this vector contains an index of the phi_related_virtuals
    /// vector, or the NO_REG value. If a cell c contains a valid index i, then
    /// it means that the collection of virtuals phi-related to virtual c is
    /// stored in phi_related_virtuals[i].
    std::vector<unsigned> v2phi;


    /// This vector associates virtuals with the phi functions in which they
    /// are defined. Notice that, although the vector will have an entry to all
    /// the virtuals, only those defined in phi-functions will be valid.
    std::vector< const MachineInstr * > def_list;


    /// This vector associates virtuals with the phi-functions in which they
    /// are used.
    std::vector< std::vector<MachineInstr *> > use_list;


    /// This vector associates virtuals to instructions. Each virtual is
    /// defined by a phi-function. def_sites[v] = {i0, i1,..., i2} is the list
    /// of instructions where the parameters of the phi-function that defines
    /// v are defined.
    std::vector< std::set<unsigned> > def_sites;


    /// This vector maps virtuals to stack slots. All the phi-related virtuals
    /// are mapped to the same stack slot.
    std::vector<int> v2s;


    /// The reference to the machine function is useful for printing and
    /// debugging:
    const MachineFunction * machine_function;


    /// runOnMachineFunction - Groups together the phi related virtuals.
    bool runOnMachineFunction(MachineFunction & mf);


    /// This method allocates the data structures.
    void initialize_data_structures(MachineFunction & mf);


    /// This method reads an instruction, and marks it as related to all the
    /// virtuals defined or used in the instruction.
    void visit_phi_function(MachineInstr * phi);


    /// This method marks the virtuals v_def and v_reg as being related. This
    /// method must be called in such a way that v_def is the register defined
    /// in the phi-function, and v_reg is the register used in the same
    /// phi-function.
    void add_to_related_set(unsigned v_def, unsigned v_reg);


    /// Free the memory, so the operating system can use it in other
    /// applications.
    void releaseMemory() {
        this->phi_related_virtuals.clear();
        this->v2phi.clear();
        this->def_list.clear();
        this->use_list.clear();
        this->def_sites.clear();
        this->v2s.clear();
    }

  };

}

#endif
