//===------------- PhiMemCoalescer_Fer.cpp - Printing prototype -----------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This pass attempts to group together virtuals whose live range does not
// overlap, so that they can be inserted in the same memory spot.
//
// Date: March 3rd, 2007
//
//===----------------------------------------------------------------------===//
//
// The combinantion of PhiSplit and PhiAnalysis passes group together all the
// virtuals that are part of the same phi-function. This is fine for
// implementing the phi-deconstruction algorithm, but it is to conservative for
// assigning the same memory address to phi-related virtuals. This pass
// attempts to increase these groups by checking if the live ranges of virtuals
// do not overlap, and then assigning the virtuals to the same group.
//
//===----------------------------------------------------------------------===//

#ifndef PHI_MEM_COALESCER_H
#define PHI_MEM_COALESCER_H

#include "llvm/PassSupport.h"

namespace llvm {

  class DefUseSites_Fer;
  class IG_Builder_Fer;
  class PhiAnalysis_Fer;
  class SplitPhiGroups_Fer;
  class EdgeLiveness_Fernando;

  class PhiMemCoalescer_Fer : public MachineFunctionPass {

    public:

    /// reporst that this pass does not change anything on the control flow
    /// graph.
    virtual void getAnalysisUsage(AnalysisUsage & AU) const;


    /// Informs the name of the pass, so it can appear in tools like llc.
    virtual const char * getPassName() const {
        return "Phi mem coalescer - Fernando.";
    }

    private:

    /// This variable contains liveness analysis information.
    EdgeLiveness_Fernando * edge_liveness;


    /// This variable maps virtual registers to the instructions where it
    /// is defined and used.
    DefUseSites_Fer * def_use_sites;


    /// This class is a interference graph. This interference graph only holds
    /// interferences between physical registers (the pre-colored registers)
    /// and virtual registers whose live ranges overlap.
    IG_Builder_Fer * ig;


    /// This class collects information about which nodes are related by phi
    /// functions, in order to allow coalescing during register allocation;
    PhiAnalysis_Fer * phi_analysis;


    /// This pass splits the live range of phi-related virtuals in such a way
    /// that they the transformed program does not contain interferences
    /// between phi-related virtuals. We need a pointer to this class because
    /// it contains the move instructions inserted to split live ranges.
    SplitPhiGroups_Fer * split_phi;


    /// runOnMachineFunction - Groups together the phi related virtuals.
    virtual bool runOnMachineFunction(MachineFunction & mf);


    /// This method initializes the data structures used during register
    /// allocation. Among the variables initilized, are ig, split_phi and
    /// def_use_sites.
    void initialize_data_structures();


    /// This method return true if any of the virtuals in these sets are
    /// adjacent in the interference graph.
    bool overlap(unsigned use_reg, unsigned def_reg);


    /// This method removes the instruction mi from the basic block, and updates
    /// the liveness analysis information and the def/use sites.
    void remove_instruction(
        MachineInstr * mi,
        unsigned def_reg,
        unsigned use_reg
    );

  };

}

#endif
