//===---------SeparatePhiLiveRanges_Fer.cpp - Printing prototype ---------===//
//
//              Register Allocation via Coloring of Chordal Graphs
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it changes the CFG to avoid live ranges of phi-related virtuals
// been alive together.
//
// Date: February 21th, 2007
//
//===----------------------------------------------------------------------===//
//
// This pass separates the live ranges of phi-related virtuals. If the basic
// block contains a construct like:
// B3 := PHI(B1, B2);
// ...
// := B3, B1 // that is, B1 and B3 are alive past the phi-function.
// Then the method inserts a move instruction to make B1, and B3 unrelated:
// V := B1;
// B3 := PHI(V, B2);
//
//===----------------------------------------------------------------------===//

#ifndef SEPARATE_PHI_LIVE_RANGES_FER_H
#define SEPARATE_PHI_LIVE_RANGES_FER_H

#include "llvm/Pass.h"

namespace llvm {

  class EdgeLiveness_Fernando;
  class DefUseSites_Fer;

  class SeparatePhiLiveRanges_Fer : public MachineFunctionPass {

    public:

    /// reporst that this pass does not change anything on the control flow
    /// graph.
    virtual void getAnalysisUsage(AnalysisUsage & AU) const;

    private:

    /// The machine_function object has references to the data structures used
    /// to produce new instruction.
    MachineFunction * machine_function;


    /// This variable contains liveness analysis information.
    EdgeLiveness_Fernando * edge_liveness;


    /// This variable maps virtual registers to the instructions where it
    /// is defined and used.
    DefUseSites_Fer * def_use_sites;


    /// This method split all the critical edges in the machine function
    /// control flow graph.
    virtual bool runOnMachineFunction(MachineFunction &mf);


    /// This method returns true if the live ranges of 'live_in_reg' and
    /// 'v_reg' overlap inside the basic block 'mbb'.
    bool overlap_in_mbb
                (unsigned live_in_reg, unsigned v_reg, MachineBasicBlock *mbb);


    /// This method creates a new virtual register with the same class as
    /// reg_model.
    unsigned create_new_virtual_register(unsigned reg_model);


    /// To raise a parameter p of a phi-function means to replace it with a new
    /// virtual v, and, in order to keep the semantics of the old program, to
    /// insert a copy v := p in the basic block that feeds this parameter to the
    /// phi-function.
    void raise_parameters(MachineInstr * phi, MachineBasicBlock * mbb);


    /// This method creates 'num_virt' new virtuals X_i, and adds 'num_virt'
    /// copy instructions such as 'X_i = v_reg' at the end of block 'mbb'. Next,
    /// it replaces each occurence of 'v_reg' as the 'op_pos'th parameter of
    /// each phi function that is finds in the vector 'phi_insts'.
    void split_live_range(
        unsigned v_reg,
        std::vector<MachineInstr *> & phi_insts,
        unsigned op_pos,
        MachineBasicBlock & mbb
    );


    /// This method inserts a move instruction at the end of basic block mbb.
    /// The move instruction copies the value of register src to register dst.
    /// The method returns the newly created instruction.
    MachineInstr * add_move(
        MachineBasicBlock & mbb,
        unsigned src,
        unsigned dst
    );

  };

}

#endif
