//===------------ SplitPhiGroups_Fer.cpp - Printing prototype -------------===//
//
//              Register Allocation via Coloring of Chordal Graphs
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it split live ranges of phi-related virtual registers.
//
// Date: February 18th, 2007
//
//===----------------------------------------------------------------------===//
//
// This pass splits the live ranges of virtual registers that contribute
// parameters to phi-functions in the same column of a phi-matrix. For example,
// if the target code contains an assignment such as:
// A1 = PHI(V, A3);
// A2 = PHI(V, A4);
// The the method will split the live range of V in two interfering live ranges
// by inserting two move instructions at the end of the basic block that feeds
// the first column of the matrix. After the method, the program will be like:
// A5 = V;
// A6 = V;
// goto ...
// A1 = PHI(A5, A3);
// A2 = PHI(A6, A4);
//
//===----------------------------------------------------------------------===//

#ifndef SPLIT_PHI_GROUPS_FER_H
#define SPLIT_PHI_GROUPS_FER_H

#include "llvm/Pass.h"

namespace llvm {

  class SplitPhiGroups_Fer : public MachineFunctionPass {

    public:

    /// reporst that this pass does not change anything on the control flow
    /// graph.
    virtual void getAnalysisUsage(AnalysisUsage & AU) const;

    /// This method prints all the ins. stored in the vector of split_points.
    void dump();

    /// Returns the vector contain all the split points.
    std::vector<MachineInstr *> & get_split_points() {
        return this->split_points;
    }

    private:

    /// The only purpose of keeping the machine function in this class is
    /// to help in the dump method.
    MachineFunction * machine_function;


    /// This variable holds the number of virtual registers created:
    unsigned num_v_regs;


    /// This vector will hold the move instructions created to partition phi
    /// functions:
    std::vector<MachineInstr *> split_points;


    /// This method split all the critical edges in the machine function
    /// control flow graph.
    virtual bool runOnMachineFunction(MachineFunction &mf);


    /// This method creates a new virtual register with the same class as
    /// reg_model.
    unsigned create_new_virtual_register(unsigned reg_model);


    /// This function takes a phi-function and the basic block where the phi
    /// function is defined, and adds one copy instruction for each parameter
    /// of the phi-function.
    void raise_parameters(MachineInstr * phi, MachineBasicBlock *mbb);


    /// This method inserts a move instruction at the end of basic block mbb.
    /// The move instruction copies the value of register src to register dst.
    /// If the source of the move is a spill, then, instead of inserting a move
    /// instruction, this method directly loads the spilled value into the
    /// destiny.
    void add_move(MachineBasicBlock & mbb, unsigned src, unsigned dst);


    /// Free the memory, so the operating system can use it in other
    /// applications.
    void releaseMemory() {
        this->split_points.clear();
    }

  };

}

#endif
