//===------------- Virtual_Fernando.cpp - Virtual Register ----------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it represents a virtual register.
//
// Date: January 29th, 2007
//
//===----------------------------------------------------------------------===//
//
// This file contains information referent to a virtual register. Such
// information is necessary during register allocation, and includes, for
// instance, the size of the live range of that virtual, the instruction where
// the virtual is defined, the instructions where it is used, etc.
//
//===----------------------------------------------------------------------===//

#ifndef VIRTUAL_FERNANDO_H
#define VIRTUAL_FERNANDO_H

namespace llvm {

    class MachineInstr;

    class Virtual_Fernando {

        public:

        ///
        Virtual_Fernando() : live_range_size(0), def(NULL)
        {
            return;
        }

        ///
        ~Virtual_Fernando() {
            uses.clear();
        }

        /// Returns the size of the live range of the virtual register.
        int get_live_range_size() const;

        /// Adds 1 to the size of the live range of this variable.
        void inc_live_range_size();

        /// Returns a textual representation of this object.
        std::string to_string() const;

        /// Returns the instruction where the temporary is defined. This
        /// instruction is unique, because the target program is expected to be
        /// in the SSA-form.
        const MachineInstr * get_def_site() const;

        /// Returns a vector containing all the instruction that use the given
        /// virtual register.
        std::vector< MachineInstr * > & get_use_sites();

        /// This instruction marks the instruction mi as the definition site of
        /// the virtual register.
        void set_def_site(const MachineInstr * mi);

        /// This method removes the instruction mi from the list of use sites
        /// of this virtual.
        void remove_use_site(MachineInstr * mi);

        /// This method adds the instruction mi as one of the use sites of the
        /// virtual.
        void add_use_site(MachineInstr * mi);

        /// This method computes the weight of this virtual. For performance
        /// reasons, the weight should be computed only once.
        const void compute_weight(LoopInfo & li);

        /// This method returns the weigth (spilling factor) of this virtual.
        unsigned get_weight() const;

        private:

        /// The weight of a virtual is a measure that helps to decide which
        /// register to spill. The bigger the weight, the more likely the
        /// virtual will not be spilled.
        unsigned weight;

        /// The size of the live range is the number of instructions that
        /// contains the virtual in its out set. 
        unsigned live_range_size;

        /// Mapping of registers to definition sites. Because in LLVM any
        /// register can be mapped to an unsigned integer, unsigned is
        /// used as the type for the keys.
        const MachineInstr * def;

        /// Mapping of registers to use sites. Contrary to definition, a
        /// virtual register can be used multiple times in the same program.
        std::vector<MachineInstr *> uses;

    };  // end of EdgeLiveness_Fernando declaration

    //===----------------------------------------------------------------------
    // Adds 1 to the size of the live range of this variable.
    //===----------------------------------------------------------------------
    inline void Virtual_Fernando::inc_live_range_size() {
        this->live_range_size++;
    }

    //===----------------------------------------------------------------------
    // Returns the size of the live range of the virtual register.
    //===----------------------------------------------------------------------
    inline int Virtual_Fernando::get_live_range_size() const {
        return this->live_range_size;
    }

    //===----------------------------------------------------------------------
    // Returns a vector containing all the instruction that use the given
    // virtual register.
    //===----------------------------------------------------------------------
    inline std::vector< MachineInstr * > & Virtual_Fernando::get_use_sites() {
        return this->uses;
    }

    //===----------------------------------------------------------------------
    // This instruction marks the instruction mi as the definition site of the
    // virtual register.
    //===----------------------------------------------------------------------
    inline void Virtual_Fernando::set_def_site(const MachineInstr * mi) {
        this->def = mi;
    }

    //===----------------------------------------------------------------------
    // This method adds the instruction mi as one of the use sites of the
    // virtual.
    //===----------------------------------------------------------------------
    inline void Virtual_Fernando::add_use_site(MachineInstr * mi) {
        this->uses.push_back(mi);
    }

    //===----------------------------------------------------------------------
    // This method returns the weigth (or spilling factor) of this virtual.
    //===----------------------------------------------------------------------
    inline unsigned Virtual_Fernando::get_weight() const {
        return this->weight;
    }

} // end llvm namespace

#endif
