//===----------- FernandoLiveStack.cpp - Printing prototype ---------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it holds the variables alive at certain point during a scan of
// the control flow graph.
//
// Date: May 28th, 2006
//
//===----------------------------------------------------------------------===//
//
// This pass collects the liveness information of the control flow graph.
// For each pair of basic blocks u and v, it records the set of variables
// alive between u and v. The requirements in terms of space is O(N*N*T),
// where N is the number of basic blocks, and T is the total number of
// temporaries in the control flow graph. Notice that this program only
// records information about virtual registers. Physical registers are
// not supposed to be alive between basic blocks.
//
//===----------------------------------------------------------------------===//

#include <vector>
#include <sstream>                               // ostringstream
#include <iostream>                              // cerr

#include "llvm/CodeGen/Passes.h"                 // EdgeLivenessFernandoPassID
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/Instructions.h"                   // MachineOperand
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include <llvm/Target/TargetMachine.h>
#include "llvm/ADT/DepthFirstIterator.h"         // df_ext_iterator

#include "llvm/fernando/CriticalEdgeRemoval_Fer.h"
#include "llvm/fernando/DefUseSites_Fer.h"
#include "llvm/fernando/EdgeLiveness_Fernando.h" // class definition

using namespace llvm;


namespace {
    RegisterPass<EdgeLiveness_Fernando>
    X("edge_liveness", "Edge liveness analysis");
}


//===--------------------------------------------------------------------------
// This method fills a vector with all the virtual registers alive between the
// origin and destiny blocks.
//===--------------------------------------------------------------------------
void EdgeLiveness_Fernando::to_vector(
    const MachineBasicBlock & org,
    const MachineBasicBlock & dst,
    std::vector<unsigned> & ans
) const {
    int row = org.getNumber();
    int col = dst.getNumber();
    const std::vector<bool> & live_variables = this->live_sets[row][col];
    ans.clear();

    for (unsigned i = 0, e = live_variables.size(); i != e; ++i) {
        if(live_variables[i]) {
            ans.push_back(i + MRegisterInfo::FirstVirtualRegister);
        }
    }
}


//===--------------------------------------------------------------------------
// returns a string in the format (t1, t2, ..., tn), where ti is one of the
// temporaries alive in the edge that connects the machine basic block
// "origin" to "destiny".
//===--------------------------------------------------------------------------
std::string EdgeLiveness_Fernando::to_string(
    const MachineBasicBlock & org,
    const MachineBasicBlock & dst
) const {
    int row = org.getNumber();
    int col = dst.getNumber();
    std::vector<bool> live_variables = this->live_sets[row][col];
    bool isFirst = true;
    std::ostringstream aux;
    aux << "(";
    for (unsigned i = 0, e = live_variables.size(); i != e; ++i) {
        if(live_variables[i]) {
            if(isFirst) {
                isFirst = false;
                aux << "t" << i;
            } else {
                aux << ", " << "t" << i;
            }
        }
    }
    aux << ")";
    return aux.str();
}


//===--------------------------------------------------------------------------
// Removes temporary from the live set of the edge connecting the block origin
// to the block destiny.
//===--------------------------------------------------------------------------
void EdgeLiveness_Fernando::kill(
                                  const MachineBasicBlock & org,
                                  const MachineBasicBlock & dst,
                                  const unsigned v_reg
                                )
{
    int row = org.getNumber();
    int col = dst.getNumber();

    if(MRegisterInfo::isVirtualRegister(v_reg)) {
        unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
        std::vector<bool> & edge = this->live_sets[row][col];
        assert(edge.size() > base_reg &&
               "EdgeLive_Fer: Index out of range in set alive.");
       edge[base_reg] = false;
    }
}


//===--------------------------------------------------------------------------
// This method informs if the given temporary is alive in any edge that leaves
// the basic block mbb.
//===--------------------------------------------------------------------------
bool EdgeLiveness_Fernando::is_live_out(
    const MachineBasicBlock & mbb,
    const unsigned v_reg
) const {
    bool res = false;
    for(MachineBasicBlock::const_succ_iterator succ_aux = mbb.succ_begin();
                                      succ_aux != mbb.succ_end(); succ_aux++) {
        const MachineBasicBlock* succ_mbb = *succ_aux;
        res = res || is_alive(mbb, *succ_mbb, v_reg);
    }
    return res;
}


//===--------------------------------------------------------------------------
// Removes the v_reg from the live set of all the blocks that can be
// reached from mbb.
//===--------------------------------------------------------------------------
void EdgeLiveness_Fernando::kill_forward
                        (const unsigned v_reg, const MachineBasicBlock & mbb) {
    for(MachineBasicBlock::const_succ_iterator succ_aux = mbb.succ_begin();
        succ_aux != mbb.succ_end(); succ_aux++) {
        const MachineBasicBlock* succ_mbb = *succ_aux;
        if( is_alive(mbb, *succ_mbb, v_reg) ) {
            kill(mbb, *succ_mbb, v_reg);
            kill_forward(v_reg, *succ_mbb);
        }
    }
}


//===--------------------------------------------------------------------------
// Removes the v_reg from the live set of all the blocks that have a
// path leading to mbb.
//===--------------------------------------------------------------------------
void EdgeLiveness_Fernando::kill_backward
                        (const unsigned v_reg, const MachineBasicBlock & mbb) {
    for(MachineBasicBlock::const_pred_iterator pred_aux = mbb.pred_begin();
                                      pred_aux != mbb.pred_end(); pred_aux++) {
        MachineBasicBlock* pred_mbb = *pred_aux;
        if( is_alive(*pred_mbb, mbb, v_reg) ) {
            kill(*pred_mbb, mbb, v_reg);
            kill_backward(v_reg, *pred_mbb);
        }
    }
}


//===--------------------------------------------------------------------------
// This method removes the live range of v_sp. This must be done after spill
// is performed. The placement of load and store instructions must be
// performed in such a way that is always safe to remove the live range of the
// spilled register.
//===--------------------------------------------------------------------------
void EdgeLiveness_Fernando::kill_live_range
                        (const unsigned v_reg, const MachineBasicBlock & mbb) {
    kill_forward(v_reg, mbb);
    kill_backward(v_reg, mbb);
}


//===--------------------------------------------------------------------------
// This method returns a vector containing the registers alive in the edge
// between the given basic blocks. Notice that each element in the returned
// vector is not the LLVM representation of the virtual register. To get the
// virtual register number, add to the MRegisterInfo::FirstVirtualRegister to
// the index in the returned vector.
//===--------------------------------------------------------------------------
const std::vector<bool> & EdgeLiveness_Fernando::get_live_variables
         (const MachineBasicBlock & org, const MachineBasicBlock & dst) const {
    int row = org.getNumber();
    int col = dst.getNumber();
    return this->live_sets[row][col];
}


//===--------------------------------------------------------------------------
// Every time a use is found, the graph is traversed backwards, and all the
// edges in which the used temporary is alive are marked.
//===--------------------------------------------------------------------------
void EdgeLiveness_Fernando::handle_use (
     const unsigned v_reg,
     const MachineBasicBlock & mbb
) {
    int block_number = mbb.getNumber();
    if(this->def_use_sites->get_def_block(v_reg) != block_number) {
        // it means that the register is not defined in this very block; thus,
        // it is alive in all the edges that reach this block, if it is not
        // used only in a PHI function.
        for (MachineBasicBlock::const_pred_iterator pred = mbb.pred_begin(),
                                       E = mbb.pred_end(); pred != E; ++pred) {
            const MachineBasicBlock *pred_block = *pred;
            if( ! is_alive(*pred_block, mbb, v_reg) ) {
                set_alive(*pred_block, mbb, v_reg);
                handle_use(v_reg, *pred_block);
            }
        }
    }
}


//===--------------------------------------------------------------------------
// This function scans every instruction of the basic block, and compute
// liveness information in a backward fashion. Every time a use is found, the
// graph is scanned backwards, and all the edges in which the used temporary is
// alive are marked. However, this does not happen if the use was found in a
// PHI instruction. In this case, only the edge from where comes the parameter
// of the PHI instruction is inspected. As a remark, PHI instruction, in LLVM,
// contain a list of operands like (v1, b1, v2, b2, ... v_n, bn), where v_i is
// a virtual register, and b_i is the block number from where the virtual
// register comes.
//===--------------------------------------------------------------------------
void EdgeLiveness_Fernando::compute_liveness(const MachineBasicBlock & block) {
    MachineBasicBlock::const_iterator mbbi = block.begin();
    while (mbbi != block.end()) {
        const MachineInstr *mi = mbbi++;
        for (unsigned i = 0; i != mi->getNumOperands(); ++i) {
            const MachineOperand & mop = mi->getOperand(i);
            if(mop.isRegister() && mop.getReg()) {
                unsigned reg = mop.getReg();
                if(MRegisterInfo::isVirtualRegister(reg) && mop.isUse()) {
                    if(mi->getOpcode() == TargetInstrInfo::PHI) {
                        const MachineBasicBlock * sbck =
                                    mi->getOperand(i+1).getMachineBasicBlock();
                        set_alive(*sbck, block, reg);
                        handle_use(reg, *sbck);
                    } else {
                        handle_use(reg, block);
                    }
                }
            }
        }
    }
}


//===--------------------------------------------------------------------------
// adds the temporary to the live set of the edge that goes from machine basic
// block "origin" to "destiny". Notice that only virtual registers are added.
// Physical registers are not supposed to be alive across basic blocks.
//===--------------------------------------------------------------------------
void EdgeLiveness_Fernando::set_alive(
                                       const MachineBasicBlock & org,
                                       const MachineBasicBlock & dst,
                                       const unsigned v_reg
) {
    int row = org.getNumber();
    int col = dst.getNumber();

    if(MRegisterInfo::isVirtualRegister(v_reg)) {
        unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
        std::vector<bool> & edge = this->live_sets[row][col];
        if(edge.size() <= base_reg) {
            edge.resize(2 * (base_reg + 1), false);
        }
        edge[base_reg] = true;
        this->add_live_range_size(1, v_reg);
    }
}


//===--------------------------------------------------------------------------
// This method adds lr_size to the current size of the live range of v_reg.
//===--------------------------------------------------------------------------
void EdgeLiveness_Fernando::add_live_range_size(
    unsigned lr_size,
    unsigned v_reg
) {
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    if(base_reg >= this->live_range_sizes.size()) {
        this->live_range_sizes.resize(2 * (base_reg + 1), 0);
    }
    this->live_range_sizes[base_reg] += lr_size;
}


//===--------------------------------------------------------------------------
// Return the size of the live range of the given virtual register. If the
// register is not a virtual, or if it is the virtual created due to spills,
// the live range size will be 0.
//===--------------------------------------------------------------------------
unsigned EdgeLiveness_Fernando::get_live_range_size(unsigned v_reg) {
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    if(base_reg >= this->live_range_sizes.size()) {
        // The virtual was created due to the spilling of another virtual.
        return 0;
    } else if( ! MRegisterInfo::isVirtualRegister(v_reg) ) {
        // v_reg is a physical register. It cannot be spilled.
        return 0;
    } else {
        return this->live_range_sizes[base_reg];
    }
}



//===--------------------------------------------------------------------------
// This method informs if the given temporary is alive in the edge between the
// blocks "origin" and "destiny".
// TODO: remove this function, and use the unsigned only.
//===--------------------------------------------------------------------------
bool EdgeLiveness_Fernando::is_alive(
                                       const MachineBasicBlock & org,
                                       const MachineBasicBlock & dst,
                                       const unsigned v_reg
) const {
    int row = org.getNumber();
    int col = dst.getNumber();
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    const std::vector<bool> & edge = this->live_sets[row][col];
    if(edge.size() <= base_reg) {
        return false;
    }
    return edge[base_reg];
}


//===--------------------------------------------------------------------------
// This method allocates memory for the data structures used in this analysis.
// These data structures comprise the matrix (O(T*N*N)) that holds liveness
// information, and the vector that maps registers to the blocks where they
// where defined (O(T)).
//===--------------------------------------------------------------------------
void EdgeLiveness_Fernando::initialize_data_structures(MachineFunction & mf) {
    unsigned n_of_blocks = mf.size();
    unsigned i, j;
    SSARegMap * srm = mf.getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    unsigned num_vreg = last_v_reg - MRegisterInfo::FirstVirtualRegister;

    // initialize the matrix that will hold data flow information.
    this->live_sets.resize(n_of_blocks);
    for(unsigned u = 0; u < n_of_blocks; u++) {
        this->live_sets[u].resize(n_of_blocks);
    }

    // gets the list of def/use sites
    this->def_use_sites = & getAnalysis<DefUseSites_Fer>();

    // Allocates space to hold the live range size of each virtual. Initially
    // we assume that all live ranges have size 0. Belive it or not, it makes
    // difference if I write resize(num_vreg, 0). The final code is a little
    // worse.
    this->live_range_sizes.resize(num_vreg);
    for(unsigned u = 0; u < this->live_range_sizes.size(); u++) {
        this->live_range_sizes[u] = 0;
    }
}


//===--------------------------------------------------------------------------
// Free the memory, so the operating system can use it in other applications.
//===--------------------------------------------------------------------------
void EdgeLiveness_Fernando::releaseMemory() {
    this->live_sets.clear();
    this->live_range_sizes.clear();
}


//===--------------------------------------------------------------------------
// The method prints all the results obtained during the liveness analysis.
// It only print liveness analysis results. Although this class also contains
// a data structure that marks the block where each temporary is defined,
// the dump method does not print this information.
//===--------------------------------------------------------------------------
void EdgeLiveness_Fernando::dump() const {
    std::cerr << "Live analysis information:\n";
    for (MachineFunction::iterator mbb = this->machine_function->begin(),
         mbbe = this->machine_function->end();
         mbb != mbbe; ++mbb) {
        for (MachineBasicBlock::const_pred_iterator pred = mbb->pred_begin(),
             E = mbb->pred_end(); pred != E; ++pred) {
            const MachineBasicBlock *pred_block = *pred;
            int row = pred_block->getNumber();
            int col = mbb->getNumber();
            std::cerr << "(" << row << ", " << col << ") "
            << to_string(*pred_block, *mbb) << "\n";
        }
    }
    std::cerr << "Live range sizes:\n";
    for(unsigned u = 0; u < this->live_range_sizes.size(); u++) {
        std::cerr << "t" << u << ": " << this->live_range_sizes[u] << "\n";
    }
}


//===--------------------------------------------------------------------------
// Collects liveness information about each edge of the control flow graph.
//===--------------------------------------------------------------------------
bool EdgeLiveness_Fernando::runOnMachineFunction(MachineFunction & mf) {
    this->machine_function = & mf;
    this->reg_info = mf.getTarget().getRegisterInfo();

    initialize_data_structures(mf);

    // traverse the control flow graph in depth first order pinpointing the
    // registers alive in each edge:
    for (MachineFunction::iterator mbb = mf.begin(), mbbe = mf.end();
                                                          mbb != mbbe; ++mbb) {
        compute_liveness(*mbb);
    }

    return false;
}
