//===----------- IG_Builder_Fer - Builds the interference graph ----------===//
//
//                     The LLVM Compiler Infrastructure
//
// Given a control flow graph, this pass builds the correcsponding interference
// graph. This file has been implemented by fernando at cs dot ucla dot edu.
//
// Date: July 10th, 2006
//
//===---------------------------------------------------------------------===//
//
// This pass builds the interference graph for a target control flow graph.
// Two nodes in the interference graph are connected if, and only if, their
// live ranges overlap in the control flow graph.
// Important: this version of the interference graph does not contains the
// full interfering pairs. It only holds interferences between physical
// registers and virtual registers. This helps in the implementation of the
// register allocation pass, and avoids using too much memory and time to
// hold the interference graph. Actually, this graph uses about one fourth
// of the memory that it would use if the full interference graph were to be
// built and stored.
//
//===---------------------------------------------------------------------===//

#include <iostream>                               // cerr

#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/Instructions.h"                   // MachineOperand
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"

#include "llvm/fernando/RegClique_Fer.h"
#include "llvm/fernando/EdgeLiveness_Fernando.h"
#include "llvm/fernando/NeatPrint_Fer.h"
#include "llvm/fernando/IG_Builder_Fer.h"

using namespace llvm;

namespace {
    RegisterPass<IG_Builder_Fer>
    X("interference_graph", "Interference graph builder");
}



//===--------------------------------------------------------------------------
// Free the memory, so the operating system can use it in other applications.
//===--------------------------------------------------------------------------
void IG_Builder_Fer::releaseMemory() {
    this->constraint_map.clear();
}


//===--------------------------------------------------------------------------
// Returns true if there is an edge linking p_reg and v_reg, and false
// otherwise.
//===--------------------------------------------------------------------------
bool IG_Builder_Fer::is_connected(unsigned p_reg, unsigned v_reg) {
/*
    assert(MRegisterInfo::isPhysicalRegister(p_reg) &&
                                           "Target register must be physical");
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                                            "Source register must be virtual");
*/
    return this->constraint_map.are_adj(p_reg, v_reg);
}


//===--------------------------------------------------------------------------
// This method prints the constraints in the standard error output.
//===--------------------------------------------------------------------------
void IG_Builder_Fer::dump_constraints() {
    this->constraint_map.dump(*this->machine_function);
}


//===--------------------------------------------------------------------------
// This method removes all the interferences in the graph.
//===--------------------------------------------------------------------------
void IG_Builder_Fer::remove_interferences() {
std::cerr << "Calling remove interferences\n";
    this->constraint_map.clear();
}


//===--------------------------------------------------------------------------
// This method marks the physical register p_reg and the virtual register
// v_reg as overlapping. Thus, the register allocator must avoid assigning to
// v_reg the color assigned to v_reg.
//===--------------------------------------------------------------------------
void IG_Builder_Fer::add_interference(unsigned p_reg, unsigned v_reg) {
    this->constraint_map.connect(v_reg, p_reg);
    // Add interferences for all the physical alias of p_reg
    if(MRegisterInfo::isPhysicalRegister(p_reg)) {
        const TargetMachine & target_machine = machine_function->getTarget();
        const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
        for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
            this->constraint_map.connect(v_reg, *a);
        }
    }
}



//===--------------------------------------------------------------------------
// Return true if mi defines v_reg, and false otherwise.
//===--------------------------------------------------------------------------
bool is_defining_instr(const MachineInstr * mi, unsigned v_reg) {
    for(unsigned u = 0; u < mi->getNumOperands(); u++) {
        const MachineOperand & mo = mi->getOperand(u);
        if(mo.isRegister() && mo.getReg() && mo.isDef()) {
            if(mo.getReg() == v_reg) {
                return true;
            }
        }
    }
    return false;
}


//===--------------------------------------------------------------------------
// This method marks v_reg as interfering with all the virtuals used in mi.
//===--------------------------------------------------------------------------
void IG_Builder_Fer::compute_interfs_with_uses(
    const MachineInstr *mi,
    unsigned v_reg
) {
    const TargetMachine & target_machine = machine_function->getTarget();
    const TargetInstrInfo & tii = * (target_machine.getInstrInfo());
    const TargetInstrDescriptor & tid = tii.get(mi->getOpcode());

    // compute interferences against implicitly used regs
    if(tid.ImplicitUses) {
        for(const unsigned *implicit_use = tid.ImplicitUses;
                                           *implicit_use; implicit_use++) {
            add_interference(*implicit_use, v_reg);
        }
    }

    // compute interferences against explicitly used regs
    for(unsigned u = 0; u < mi->getNumOperands(); u++) {
        if(mi->getOpcode() == TargetInstrInfo::PHI) {
            break;
        }
        const MachineOperand & mo = mi->getOperand(u);
        if(mo.isRegister() && mo.isUse() && mo.getReg()) {
            add_interference(mo.getReg(), v_reg);
        }
    }
}


//===--------------------------------------------------------------------------
// This method marks v_reg as interfering with all the virtuals defined in mi.
//===--------------------------------------------------------------------------
void IG_Builder_Fer::compute_interfs_with_defs(
    const MachineInstr *mi,
    unsigned v_reg
) {
    const TargetMachine & target_machine = machine_function->getTarget();
    const TargetInstrInfo & tii = * (target_machine.getInstrInfo());
    const TargetInstrDescriptor & tid = tii.get(mi->getOpcode());

    // compute interferences against implicitly defined regs
    if(tid.ImplicitDefs) {
        for (const unsigned *implicit_def = tid.ImplicitDefs;
                                           *implicit_def; implicit_def++) {
            this->add_interference(*implicit_def, v_reg);
        }
    }

    // compute interferences against explicitly defined regs
    for(unsigned u = 0; u < mi->getNumOperands(); u++) {
        const MachineOperand & mo = mi->getOperand(u);
        if(mo.isRegister() && mo.isDef() && mo.getReg()) {
            add_interference(mo.getReg(), v_reg);
        }
    }
}


//===--------------------------------------------------------------------------
// This method scans the live range of v_reg, adding to the interference graph
// all the interferences found between v_reg and any physical register that is
// alive between the instructions cur and begin.
//===--------------------------------------------------------------------------
unsigned IG_Builder_Fer::find_interferences(
    MachineBasicBlock::const_iterator cur,
    MachineBasicBlock::const_iterator & bgn,
    unsigned v_reg
) {
    const MachineInstr * mi;
    unsigned live_range_size = 0;

    while(true) {
        mi = cur;
        if(is_defining_instr(mi, v_reg)) {
            return live_range_size;
        }
        this->compute_interfs_with_uses(mi, v_reg);

        // if this is the first instruction in the basic block, it is over.
        if(cur == bgn) {
            return live_range_size;
        } else {
            cur--;
            live_range_size++;
        }

        mi = cur;

        // This is just a hammer to avoid that the live range be counted
        // between phi-functions.
        if(mi->getOpcode() == TargetInstrInfo::PHI) {
            live_range_size--;
        }

        this->compute_interfs_with_defs(mi, v_reg);
    }
}


//===--------------------------------------------------------------------------
// This method sets the constraints on the mapping between physical registers
// and virtual registers. A constraint between two registers, e.g.: p_reg and
// v_reg is an indication that the live range of v_reg overlaps with the live
// range of a pre-colored register p_reg.
//===--------------------------------------------------------------------------
void IG_Builder_Fer::find_interferences
                          (RegClique_Fer & ql, const MachineBasicBlock & mbb) {

    MachineBasicBlock::const_iterator mbb_cur = mbb.end();
    MachineBasicBlock::const_iterator mbb_bgn = mbb.begin();
    if(mbb_cur == mbb_bgn) {
        return;
    } else {
        mbb_cur--;
    }

    // First, scan the live ranges of the virtuals alive at the end of the
    // basic block:
    const MachineFunction * mf = mbb.getParent();
    const SSARegMap * srm = mf->getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    for(
        unsigned v_reg = MRegisterInfo::FirstVirtualRegister;
        v_reg < last_v_reg;
        v_reg++
    ) {
        if(ql.contains_reg(v_reg)) {
            const MachineInstr * mi = mbb_cur;
            if(!is_defining_instr(mi, v_reg)) {
                this->compute_interfs_with_defs(mi, v_reg);
            }
            unsigned lr_size = find_interferences(mbb_cur, mbb_bgn, v_reg);
            this->edge_liveness->add_live_range_size(lr_size, v_reg);
        }
    }

    // Now, for each instruction, every time it is used, scans its live range
    // until the beginning of the basic block:
    mbb_cur = mbb.end();
    while (mbb_cur != mbb_bgn) {
        mbb_cur--;
        const MachineInstr * mi = mbb_cur;
        if(mi->getOpcode() == TargetInstrInfo::PHI) {
            break;
        }
        for(unsigned u = 0; u < mi->getNumOperands(); u++) {
            const MachineOperand & mo = mi->getOperand(u);
            if(mo.isRegister() && mo.getReg() && mo.isUse()) {
                if(MRegisterInfo::isVirtualRegister(mo.getReg())) {
                    if(!ql.contains_reg(mo.getReg())) {
                        unsigned l_size =
                             find_interferences(mbb_cur, mbb_bgn, mo.getReg());
                        edge_liveness->add_live_range_size(l_size, mo.getReg());
                        ql.add_reg(mo.getReg());
                    }
                }
            }
        }
    }
}


//===--------------------------------------------------------------------------
// This method modifies the cliques in the cl vector to reflect the newly
// created virtual registers. Such virtuals have been created due to spilling.
//===--------------------------------------------------------------------------
void IG_Builder_Fer::rebuild_list_cliques
(unsigned sp, std::vector<RegClique_Fer *> &cl, const MachineBasicBlock &mbb) {

    // Augment each clique, so they can support the newly created virtual
    // registers.
    const MachineFunction *mf = mbb.getParent();
    SSARegMap * srm = mf->getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    for(unsigned u = 0; u < cl.size(); u++) {
        cl[u]->grow_virtual(last_v_reg - MRegisterInfo::FirstVirtualRegister);
    }

    MachineBasicBlock::const_iterator mbbi = mbb.end();
    const MachineInstr * mi;
    unsigned clique_counter = 1;

    while (mbbi != mbb.begin()) {
        mbbi--;
        mi = mbbi;
        bool isPhi = (mi->getOpcode() == TargetInstrInfo::PHI) ? true : false ;

        // Remove the virtual registers that only represent spills:
        RegClique_Fer * clique = cl[clique_counter];
        std::vector<unsigned> contents;
        clique->to_vector(contents);
        for(unsigned i = 0; i < contents.size(); i++) {
            if(contents[i] > sp) {
                clique->remove_reg(contents[i]);
            }
        }

        // add the new uses (some of the old uses will be added too, but it
        // does not matter).
        for(unsigned i = 0; i != mi->getNumOperands(); i++) {
            const MachineOperand & mo = mi->getOperand(i);
            if (mo.isUse() && mo.isRegister() && mo.getReg() && (!isPhi) ) {
                clique->add_reg(mo.getReg());
            }
        }

        clique_counter++;
    }
}


//===--------------------------------------------------------------------------
// This method updates the set of registers with data from the given
// instruction. Any defined register is removed from the clique, and any used
// register is added to the clique. If the used register is already in the
// clique, nothing is done. The value sp_cut represents the highest virtual
// register that is not source of a spill operation. Spilled virtuals must not
// be alive in more than one clique.
//===--------------------------------------------------------------------------
RegClique_Fer * IG_Builder_Fer::update_list_of_cliques(
    unsigned sp_cut,
    const MachineBasicBlock & mbb,
    RegClique_Fer & q,
    const MachineInstr & mi
) {
    RegClique_Fer * ret_value = new RegClique_Fer(q);
    ret_value->remove_definitions(mi);
    // Remove the values that represent spilled registers. They have very
    // short live ranges, only one clique wide.
    std::vector<unsigned> regs;
    ret_value->to_vector(regs);
    for(unsigned u = 0; u < regs.size(); u++) {
        if(regs[u] > sp_cut) {
            ret_value->remove_reg(regs[u]);
        }
    }
    ret_value->add_uses(mi);
    return ret_value;
}


//===--------------------------------------------------------------------------
// If liveness analysis information is available, the set of cliques in a basic
// block can be found with a backwards linear scan. The first clique will be
// the out set of the basic block. Each time a new instruction is scanned, the
// register that it defines is removed from the clique, and any register that
// it uses, if not already in the clique, is added to it. Notice that phi
// instructions do not contribute uses for clique construction. Only registers
// defined in the phi function are taken into consideration. The value sp_cut
// represents the highest virtual register that is not source of a spill
// operation. Spilled virtuals must not be alive in more than one clique.
//===--------------------------------------------------------------------------
void IG_Builder_Fer::build_list_cliques(
    unsigned sp_cut,
    std::vector<RegClique_Fer *> & cl,
    const MachineBasicBlock & mbb
) {
    RegClique_Fer * q = new RegClique_Fer
                           (*this->machine_function, mbb, this->edge_liveness);
    cl.push_back( q );

    MachineBasicBlock::const_iterator mbbi = mbb.end();
    const MachineInstr * mi;
    unsigned i = 0;
    while (mbbi != mbb.begin()) {
        mbbi--;
        mi = mbbi;
        cl.push_back( update_list_of_cliques(sp_cut, mbb, *cl[i], *mi) );
        i++;
    }
}


//===--------------------------------------------------------------------------
// This method builds the interference graph between physical and virtual
// registers.
//===--------------------------------------------------------------------------
bool IG_Builder_Fer::runOnMachineFunction(MachineFunction &mf) {

    this->machine_function = & mf;

    // gets the liveness analysis information
    this->edge_liveness = & getAnalysis<EdgeLiveness_Fernando>();

    for (MachineFunction::iterator mbb = mf.begin(), mbbe = mf.end();
                                                          mbb != mbbe; ++mbb) {
        RegClique_Fer ql(mf, *mbb, this->edge_liveness);
        find_interferences(ql, *mbb);
    }

//dump_constraints();

    return false;

}
