//===-------------- PhiAnalysis_Fer.cpp - Printing prototype --------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it groups virtual registers that are related by phi functions.
//
// Date: August 1th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class groups together virtual registers that are related by phi
// functions in the control flow graph. Such virtuals describe the same
// variable in the original program. A good register allocator should try to
// assign the same physical register to these virtuals. For instance, if
// v1 := PHI (v2, v3, v3), then {v1, v2, v3} are phi related virtuals.
//
//===----------------------------------------------------------------------===//

#include <vector>
#include <sstream>                               // ostringstream
#include <iostream>

#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/CodeGen/MachineFunctionPass.h"     // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"

#include "llvm/fernando/RegMapping_Fer.h"
#include "llvm/fernando/SplitPhiGroups_Fer.h"
#include "llvm/fernando/NeatPrint_Fer.h"
#include "llvm/fernando/PhiAnalysis_Fer.h"      // class definition


using namespace llvm;

namespace {
    RegisterPass<PhiAnalysis_Fer>
    X("PhiAnalysis", "Groups phi-related virtuals - Fernando.", true);
}

#define UNVISITED 2147483647

#define COLOR_WEIGHT 1000

#define SIMILAR_LOC_WEIGHT 1


//===--------------------------------------------------------------------------
// reporst that this pass does not change anything on the control flow graph.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::getAnalysisUsage(AnalysisUsage & AU) const {
    AU.addRequired<DefUseSites_Fer>();
    AU.addRequired<SplitPhiGroups_Fer>();
    AU.setPreservesAll();
}


//===--------------------------------------------------------------------------
// This method returns true if the given virtual (or, for the same effect,
// any other phi-related virtual) has been mapped to a stack slot.
//===--------------------------------------------------------------------------
bool PhiAnalysis_Fer::has_stack_slot(unsigned v) {
    unsigned base_reg = v - MRegisterInfo::FirstVirtualRegister;
    if(base_reg >= this->v2s.size()) {
        return false;
    } else {
        return this->v2s[base_reg] != -1;
    }
}


//===--------------------------------------------------------------------------
// This method returns the stack slot that has been assigned to virtual
// v, if it has been spilled to memory. All the phi-related virtuals have
// the same stack slot.
//===--------------------------------------------------------------------------
int PhiAnalysis_Fer::get_stack_slot(unsigned v) {
    unsigned base_reg = v - MRegisterInfo::FirstVirtualRegister;
    if(base_reg >= this->v2s.size()) {
        return -1;
    } else {
        return this->v2s[base_reg];
    }
}


//===--------------------------------------------------------------------------
// This method maps the virtual v to the stack slot s. It also has the
// effect of assigning s to all other virtuals that are phi-related to v.
// If v has already been mapped to a stack slot, the function will produce
// an error.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::map_to_stack_slot(unsigned v, int s) {
    unsigned base_reg = v - MRegisterInfo::FirstVirtualRegister;
    assert(base_reg < this->v2s.size() &&
                      "Index of register out of bounds of stack slot vector.");
    assert(s >= 0 && "Invalid stack slot.");
    assert(this->v2s[base_reg] == -1 &&
                     "Error: virtual has already been assigned a stack slot.");
    this->v2s[base_reg] = s;
    // If the virtual is related to other variables, map them all to the same
    // stack slot:
    if(this->v2phi[base_reg] != UNVISITED) {
        std::set<unsigned> * phi_set = phi_related_virtuals[v2phi[base_reg]];
        std::set<unsigned>::iterator pos;
        for (pos = phi_set->begin(); pos != phi_set->end(); ++pos) {
            unsigned r_src = * pos;
            unsigned b_reg = r_src - MRegisterInfo::FirstVirtualRegister;
            this->v2s[b_reg] = s;
        }
    }
}


/*
 * This is an implementation of shell-sort, used to sort the vector of color
 * preferences. This function gets two vectors, one containing the weight of
 * each color, and another containing the indices of the colors.
 */
void shellsort(std::vector<unsigned> & nums, std::vector<unsigned> & idxs) {
    unsigned i, j, k, tmp, tmp_idx;
    k = 3;
    while( k > 0 ) {
        for( i=0; i < nums.size(); i++ ) {
            j = i;
            tmp = nums[ i ];
            tmp_idx = idxs[ i ];
            while(( j >= k ) && ( nums[ j - k ] <= tmp )) {
                nums[ j ] = nums[ j - k ];
                idxs[ j ] = idxs[ j - k ];
                j = j - k;
            }
            nums[ j ] = tmp;
            idxs[ j ] = tmp_idx;
        }
        if(( k / 2 ) != 0) {
            k = ( k / 2 );
        } else {
            if ( k == 1 ) {
                k = 0;
            } else {
                k = 1;
            }
        }
    }
}


//===--------------------------------------------------------------------------
// This method returns all the phi-function in which the virtual v_reg is used.
//===--------------------------------------------------------------------------
std::vector<MachineInstr *> & PhiAnalysis_Fer::get_use_list(unsigned v_reg) {
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    assert(this->use_list.size() > base_reg &&
                                   "v_reg has not been found by phi-analysis");
    return this->use_list[base_reg];
}


//===--------------------------------------------------------------------------
// This method returns a vector {c_0, c_1, ..., c_n-1}, where each cell
// is a color that best fits v_reg for register allocation purposes. This
// means that, if v_reg is assigned c_0, more phi-function entries will be
// coalesced than if v_reg is assigned c_1 and so on.
//===--------------------------------------------------------------------------
const std::vector<unsigned> * PhiAnalysis_Fer::get_preferences(
    unsigned v_reg,
    const RegMapping_Fer * rmap
) {
    std::vector<unsigned> vec;
    vec.resize(this->num_p_regs, 0);
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;

    // compute preferences for the phi-function that defines v_reg
    const MachineInstr * phi_def = this->def_list[base_reg];
    if(phi_def != NULL) {
        for(unsigned i = phi_def->getNumOperands() - 1; i >= 2; i -= 2) {
            unsigned use_reg = phi_def->getOperand(i-1).getReg();
            unsigned location = rmap->get_physical_location(use_reg);
            if(location != RegMapping_Fer::NO_REG) {
                vec[location] += COLOR_WEIGHT;
            }
        }
    }
   
    // compute preferences for phi-functions that use v_reg
    for(unsigned u = 0; u < this->use_list[base_reg].size(); u++) {
        const MachineInstr * phi_use = this->use_list[base_reg][u];
        unsigned def_reg = phi_use->getOperand(0).getReg();
        unsigned def_loc = rmap->get_physical_location(def_reg);
        if(def_loc != RegMapping_Fer::NO_REG) {
            vec[def_loc] += COLOR_WEIGHT;
        }
        for (unsigned i = phi_use->getNumOperands() - 1; i >= 2; i -= 2) {
            unsigned use_reg = phi_use->getOperand(i-1).getReg();
            unsigned location = rmap->get_physical_location(use_reg);
            if(location != RegMapping_Fer::NO_REG) {
                vec[location] += SIMILAR_LOC_WEIGHT;
            }
        }
    }

    std::vector<unsigned> * colors;
    colors = new std::vector<unsigned>(this->num_p_regs, 0);
    for(unsigned u = 0; u < colors->size(); u++) {
        (*colors)[u] = u;
    }
    shellsort(vec, *colors);
    unsigned max_color = 0;
    for(unsigned u = 0; u < vec.size(); u++) {
        if(vec[u] > 0) {
            max_color++;
        }
    }
    for(unsigned u = max_color; u < colors->size(); u++) {
        (*colors)[u] = 0;
    }
    return colors;
}


//===--------------------------------------------------------------------------
// This method returns all the definition sites of the given virtual. An
// error will be produced if the virtual is not in the definition list of
// a phi-equation.
//===--------------------------------------------------------------------------
std::set<unsigned> & PhiAnalysis_Fer::get_def_sites(unsigned v_reg) {
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    assert(this->def_list[base_reg] != NULL &&
                    "The virtual is not in the definition of a phi function.");
    return this->def_sites[base_reg];
}


//===--------------------------------------------------------------------------
// This function computes the difference s - d = r, and inserts r into d. If
// r is not empty, it will return true, and will return false otherwise.
//===--------------------------------------------------------------------------
bool add_extra_regs(std::set<unsigned> & d, std::set<unsigned> & s) {
    bool aux = false;
    std::set<unsigned>::iterator pos;
    for (pos = s.begin(); pos != s.end(); ++pos) {
        unsigned reg_src = *pos;
        if(d.find(reg_src) == d.end()) {
            d.insert(reg_src);
            aux = true;
        }
    }
    return aux;
}


//===--------------------------------------------------------------------------
// This method finds all the definition sites for phi-functions. Given a
// phi-function i, the set of its definition sites is formed by instructions
// that define the parameters of i. The algorithm is iterative, because there
// may be recursive definitions between phi functions.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::find_definition_sites(DefUseSites_Fer *du) {
    bool modified;
    this->def_sites.resize(this->max_virtual + 1);
    do {
        modified = false;
        for(unsigned u = 0; u < this->max_virtual; u++) {
            if(this->def_list[u] != NULL) {
                unsigned v_reg = u + MRegisterInfo::FirstVirtualRegister;
                MachineInstr * phi = du->get_def_site(v_reg);
                for (unsigned i = phi->getNumOperands() - 1; i >= 2; i -= 2) {
                    unsigned use_reg = phi->getOperand(i-1).getReg();
                    MachineInstr * def = du->get_def_site(use_reg);
                    if(def->getOpcode() != TargetInstrInfo::PHI) {
                        if(def_sites[u].find(use_reg) == def_sites[u].end()) {
                            def_sites[u].insert(use_reg);
                            modified = true;
                        }
                    } else {
                        unsigned new_def = def->getOperand(0).getReg();
                        unsigned b_new = new_def -
                                           MRegisterInfo::FirstVirtualRegister;
                        bool r = add_extra_regs(def_sites[u], def_sites[b_new]);
                        modified |= r;
                    }
                }
            }
        }
    } while (modified == true);
}


//===--------------------------------------------------------------------------
// This method marks the virtuals v_def and v_reg as being related. This
// method must be called in such a way that v_def is the register defined in
// the phi-function, and v_reg is the register used in the same phi-function.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::add_to_related_set(unsigned v_def, unsigned v_reg) {
    unsigned bv = v_reg - MRegisterInfo::FirstVirtualRegister;
    unsigned bd = v_def - MRegisterInfo::FirstVirtualRegister;
    if(this->v2phi[bd] == UNVISITED) {
        assert(v_def == v_reg && "Lost entry in phi-collector.");
        std::set<unsigned> * set = new std::set<unsigned>();
        this->phi_related_virtuals.push_back(set);
        unsigned p_size = this->phi_related_virtuals.size();
        this->v2phi[bd] = p_size - 1;
        set->insert(v_def);
    } else {
        if(this->v2phi[bv] == UNVISITED) {
            std::set<unsigned> * set = phi_related_virtuals[v2phi[bd]];
            set->insert(v_reg);
            this->v2phi[bv] = this->v2phi[bd];
        } else {
            // merge the sets
            std::set<unsigned> * set_old = phi_related_virtuals[v2phi[bv]];
            std::set<unsigned> * set_new = phi_related_virtuals[v2phi[bd]];
            std::set<unsigned>::iterator pos;
            for (pos = set_new->begin(); pos != set_new->end(); ++pos) {
                unsigned r_src = * pos;
                unsigned b_aux = r_src - MRegisterInfo::FirstVirtualRegister;
                this->v2phi[b_aux] = this->v2phi[bv];
                set_old->insert(r_src);
            }
            this->v2phi[bd] = this->v2phi[bv];
        }
    }
}


//===--------------------------------------------------------------------------
// This method returns the set of virtuals that are phi-related to v_reg.
//===--------------------------------------------------------------------------
std::set<unsigned> & PhiAnalysis_Fer::get_phi_group(unsigned v_reg) {
    unsigned b_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    assert(this->v2phi[b_reg] != UNVISITED &&
                    "Attempt to grab the group of a non-phi-related virtual.");
    return * this->phi_related_virtuals[this->v2phi[b_reg]];
}


//===--------------------------------------------------------------------------
// This method reads an instruction, and marks it as related to all the
// virtuals defined or used in the instruction.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::visit_phi_function(MachineInstr * phi) {
    assert(phi->getOpcode() == TargetInstrInfo::PHI &&
                             "Only phi instructions define related virtuals.");

    unsigned v_def;

    // This set stores the already seen registers, so that they don't get
    // processed twice.
    std::set<unsigned> seen_regs;

    for(unsigned i = 0; i != phi->getNumOperands(); i++) {
        MachineOperand & mo = phi->getOperand(i);
        if (mo.isRegister() && mo.getReg()) {
            unsigned v_reg = mo.getReg() - MRegisterInfo::FirstVirtualRegister;
            if(v_reg >= this->use_list.size()) {
                this->use_list.resize(2 * (1 + v_reg) );
                this->def_list.resize(2 * (1 + v_reg) );
                this->v2phi.resize(2 * (1 + v_reg), UNVISITED);
            }
            if(v_reg > this->max_virtual) {
                this->max_virtual = v_reg;
            }
            if(mo.isDef()) {
                v_def = mo.getReg();
                add_to_related_set(v_def, v_def);
                this->def_list[v_reg] = phi;
            } else {
                if(seen_regs.find(v_reg) != seen_regs.end()) {
                    continue;
                } else {
                    add_to_related_set(v_def, mo.getReg());
                    seen_regs.insert(v_reg);
                    this->use_list[v_reg].push_back(phi);
                }
            }
        }
    }
}


//===--------------------------------------------------------------------------
// This method returns true if v_reg is part of a phi function somewhere
// in the control flow graph, and false otherwise. 
//===--------------------------------------------------------------------------
bool PhiAnalysis_Fer::is_phi_related(unsigned v_reg) {
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                       "Only virtual registers can be part of phi-functions.");
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    if(base_reg > this->max_virtual) {
        return false;
    }
    return (use_list[base_reg].size() > 0) || (def_list[base_reg] != NULL);
}


//===--------------------------------------------------------------------------
// Prints the set of phi-related virtuals.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::dump() {

    std::cerr << "Def list:\n";
    for(unsigned u = 0; u < this->max_virtual; u++) {
        if(this->def_list[u] != NULL) {
            std::cerr << "t" << u << " is defined in phi-function.\n";
        }
    }

    std::cerr << "Use list:\n";
    for(unsigned u = 0; u < this->max_virtual; u++) {
        std::cerr << "t" << u << ") " << this->use_list[u].size() << "\n";
    }

    std::cerr << "Definition sites:\n";
    for(unsigned u = 0; u < this->def_sites.size(); u++) {
        if(this->def_sites[u].size() > 0) {
            std::set<unsigned> & s = def_sites[u];
            std::set<unsigned>::iterator pos;
            std::cerr << "t" << u << ")\n";
            for (pos = s.begin(); pos != s.end(); ++pos) {
                unsigned r_src = * pos;
                unsigned b_reg = r_src - MRegisterInfo::FirstVirtualRegister;
                std::cerr << "t" << b_reg << ", ";
            }
            std::cerr << "\n";
        }
    }

    std::cerr << "Phi Groups:\n";
    for(unsigned u = 0; u < this->phi_related_virtuals.size(); u++) {
        const std::set<unsigned> * s = this->phi_related_virtuals[u];
        std::set<unsigned>::iterator pos;
        std::cerr << "Group " << u << ": (";
        for (pos = s->begin(); pos != s->end(); ++pos) {
            unsigned r_src = * pos;
            unsigned b_reg = r_src - MRegisterInfo::FirstVirtualRegister;
            std::cerr << "t" << b_reg << "  ";
        }
        std::cerr << ")\n";
    }

    std::cerr << "Phi Related Virtuals:\n";
    for(unsigned u = 0; u < this->max_virtual; u++) {
        std::cerr << u << " -> " << this->v2phi[u] << "\n";
    }
}


//===--------------------------------------------------------------------------
// This method allocates memory for the data structures.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::initialize_data_structures(MachineFunction & mf) {
    DefUseSites_Fer * def_use_sites = & getAnalysis<DefUseSites_Fer>();

    this->max_virtual = 0;
    this->machine_function = & mf;

    // Finds the number of physical registers:
    const TargetMachine & target_machine = mf.getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    this->num_p_regs = reg_info->getNumRegs();

    // Finds the number of virtual registers:
    SSARegMap * srm = mf.getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    unsigned num_v_regs = last_v_reg - MRegisterInfo::FirstVirtualRegister;
    // This size must generally be enough for all the virtuals plus the regs
    // that can be created due to spills.
    this->use_list.resize(2 * (num_v_regs + 1));
    this->def_list.resize(2 * (num_v_regs + 1));
    this->v2phi.resize(2 * (num_v_regs + 1), UNVISITED);
}


//===--------------------------------------------------------------------------
// This function merges the phi-group of v1 with the phi-group of v2, so that
// each virtual in any of these groups will be phi-related.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::merge(unsigned v1, unsigned v2) {
    unsigned b1 = v1 - MRegisterInfo::FirstVirtualRegister;
    unsigned b2 = v2 - MRegisterInfo::FirstVirtualRegister;
    assert( (this->v2phi[b1] != UNVISITED || this->v2phi[b2] != UNVISITED) &&
                          "At least one of the virtuals must be phi-related.");
    if(this->v2phi[b1] == UNVISITED) {
        std::set<unsigned> * set = phi_related_virtuals[v2phi[b2]];
        set->insert(v1);
        this->v2phi[b1] = this->v2phi[b2];
    } else if (this->v2phi[b2] == UNVISITED) {
        std::set<unsigned> * set = phi_related_virtuals[v2phi[b1]];
        set->insert(v2);
        this->v2phi[b2] = this->v2phi[b1];
    } else {
        // merge the sets
        std::set<unsigned> * set1 = phi_related_virtuals[v2phi[b1]];
        std::set<unsigned> * set2 = phi_related_virtuals[v2phi[b2]];
        std::set<unsigned>::iterator pos;
        for (pos = set2->begin(); pos != set2->end(); ++pos) {
            unsigned r_src = * pos;
            unsigned b_aux = r_src - MRegisterInfo::FirstVirtualRegister;
            this->v2phi[b_aux] = this->v2phi[b1];
            set1->insert(r_src);
        }
    }
}


//===--------------------------------------------------------------------------
// runOnMachineFunction - Groups together the phi related virtuals.
//===--------------------------------------------------------------------------
bool PhiAnalysis_Fer::runOnMachineFunction(MachineFunction & mf) {

    this->initialize_data_structures(mf);

    for (MachineFunction::iterator mbbi = mf.begin(), mbbe = mf.end();
                                                        mbbi != mbbe; ++mbbi) {
        MachineBasicBlock * mbb = mbbi;
        MachineBasicBlock::iterator mii = mbb->begin();
        while (mii != mbb->end()) {
            MachineInstr * mi = mii;
            if(mi->getOpcode() == TargetInstrInfo::PHI) {
                visit_phi_function(mi);
            } else {
                // Always safe, because phi functions are always the first
                // instructions in the basic block.
                break;
            }
            mii++;
        }
    }

    // Reserve space for the mappings of virtuals to stack slots:
    this->v2s.resize(this->v2phi.size(), -1);

//NeatPrint_Fer::print_phi_graph(mf);
//dump();

    return false;
}
