//===------------- PhiMemCoalescer_Fer.cpp - Printing prototype -----------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This pass attempts to group together virtuals whose live range does not
// overlap, so that they can be inserted in the same memory spot.
//
// Date: March 3rd, 2007
//
//===----------------------------------------------------------------------===//
//
// The combinantion of PhiSplit and PhiAnalysis passes group together all the
// virtuals that are part of the same phi-function. This is fine for
// implementing the phi-deconstruction algorithm, but it is to conservative for
// assigning the same memory address to phi-related virtuals. This pass
// attempts to increase these groups by checking if the live ranges of virtuals
// do not overlap, and then assigning the virtuals to the same group.
//
//===----------------------------------------------------------------------===//

#include <iostream>                                      // cerr

#include "llvm/CodeGen/MachineFunctionPass.h"            // MachineBasicBlock
#include "llvm/Instructions.h"                   // MachineOperand, etc
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"

#include "llvm/fernando/NeatPrint_Fer.h"
#include "llvm/fernando/EdgeLiveness_Fernando.h"
#include "llvm/fernando/IG_Builder_Fer.h"
#include "llvm/fernando/DefUseSites_Fer.h"
#include "llvm/fernando/SplitPhiGroups_Fer.h"
#include "llvm/fernando/PhiAnalysis_Fer.h"
#include "llvm/fernando/PhiMemCoalescer_Fer.h"      // class definition


using namespace llvm;

namespace {
    RegisterPass<PhiMemCoalescer_Fer>
    X("PhiMemCoalescer", "Coalesce phi-related virtuals - Fernando.", true);
}


//===--------------------------------------------------------------------------
// This pass does not change anything on the control flow graph, but it needs
// DefUseSites, IG_Builder, PhiAnalysis and SplitPhiGroups.
//===--------------------------------------------------------------------------
void PhiMemCoalescer_Fer::getAnalysisUsage(AnalysisUsage & AU) const {
    AU.addRequired<EdgeLiveness_Fernando>();
    AU.addRequired<PhiAnalysis_Fer>();
    AU.addRequired<IG_Builder_Fer>();
    AU.addRequired<SplitPhiGroups_Fer>();
    AU.addRequired<DefUseSites_Fer>();
    AU.setPreservesAll();
}


//===--------------------------------------------------------------------------
// This method initializes the data structures used during register allocation.
// Among the variables initilized, are ig, split_phi and def_use_sites.
//===--------------------------------------------------------------------------
void PhiMemCoalescer_Fer::initialize_data_structures() {
    // gets the liveness analysis information
    this->edge_liveness = & getAnalysis<EdgeLiveness_Fernando>();

    // gets the list of def/use sites
    this->def_use_sites = & getAnalysis<DefUseSites_Fer>();

    // builds the interference graph. Notice that this interference graph
    // only holds interferences between physical and virtual registers.
    this->ig = & getAnalysis<IG_Builder_Fer>();

    // Gets information about how virtuals are related to phi-functions.
    this->phi_analysis = & getAnalysis<PhiAnalysis_Fer>();

    // The colection of instructions inserted to split live ranges.
    this->split_phi = & getAnalysis<SplitPhiGroups_Fer>();
}


//===--------------------------------------------------------------------------
// This method return true if any of the virtuals in these sets are adjacent
// in the interference graph.
//===--------------------------------------------------------------------------
bool PhiMemCoalescer_Fer::overlap(unsigned use_reg, unsigned def_reg) {
    std::set<unsigned> aux_set;
    std::set<unsigned> & su = aux_set;
    if(this->phi_analysis->is_phi_related(use_reg)) {
        su = this->phi_analysis->get_phi_group(use_reg);
    } else {
        aux_set.insert(use_reg);
    }
    std::set<unsigned> & sd = this->phi_analysis->get_phi_group(def_reg);

    // Find the interferences: quadratic time:
    std::set<unsigned>::iterator pos_u;
    for (pos_u = su.begin(); pos_u != su.end(); ++pos_u) {
        unsigned vu = * pos_u;
        std::set<unsigned>::iterator pos_d;
        for (pos_d = sd.begin(); pos_d != sd.end(); ++pos_d) {
            unsigned vd = * pos_d;
            if(this->ig->is_connected(vu, vd)) {
                return true;
            }
        }
    }

    return false;
}


//===--------------------------------------------------------------------------
// This method removes the instruction mi from the basic block, and updates
// the liveness analysis information and the def/use sites.
//===--------------------------------------------------------------------------
void PhiMemCoalescer_Fer::remove_instruction(
    MachineInstr * mi,
    unsigned def_reg,
    unsigned use_reg
) {
    // find the basic block
    MachineBasicBlock * mbb = mi->getParent();

    // Get the phi-function where def_reg is used:
    std::vector<MachineInstr *> & vec = phi_analysis->get_use_list(def_reg);
    assert(vec.size() == 1 && "Split virtual is not used exactly once.");
    MachineInstr * phi = vec[0];

    // rewrite use sites of use_reg:
    this->def_use_sites->remove_use_site(mi, use_reg);
    this->def_use_sites->add_use_site(phi, use_reg);

    // Update the liveness analysis information:
    bool update = false;
    for(MachineBasicBlock::const_succ_iterator succ_aux = mbb->succ_begin();
                                     succ_aux != mbb->succ_end(); succ_aux++) {
        const MachineBasicBlock* succ_mbb = *succ_aux;
        if(this->edge_liveness->is_alive(*mbb, *succ_mbb, def_reg)) {
            update = true;
            this->edge_liveness->kill(*mbb, *succ_mbb, def_reg);
            this->edge_liveness->set_alive(*mbb, *succ_mbb, use_reg);
            break;
        }
    }
    assert(update && "Information not update in control flow graph.");

    // Update the use of the register in the phi-function
    for(unsigned u = 0; u < phi->getNumOperands(); u++) {
        MachineOperand & mo = phi->getOperand(u);
        if(mo.isRegister() && mo.getReg() == def_reg) {
            mo.setReg(use_reg);
            break;
        }
    }

    // delete the move instruction:
    //mbb->erase(mi);
    mi->eraseFromParent();
}


//===--------------------------------------------------------------------------
// runOnMachineFunction - Groups together the phi related virtuals.
//===--------------------------------------------------------------------------
bool PhiMemCoalescer_Fer::runOnMachineFunction(MachineFunction & mf) {
    const TargetMachine & target_machine = mf.getTarget();
    const TargetInstrInfo & tii = * (target_machine.getInstrInfo());
    this->initialize_data_structures();

    const std::vector<MachineInstr *> & split_points =
                                           this->split_phi->get_split_points();
    for(unsigned u = 0; u < split_points.size(); u++) {
        MachineInstr * mi = split_points[u];
        unsigned use_reg, def_reg;
        if(tii.isMoveInstr(*mi, use_reg, def_reg)) {
            if( ! this->overlap(use_reg, def_reg) ) {
                this->phi_analysis->merge(use_reg, def_reg);
                remove_instruction(mi, def_reg, use_reg);
            }
        }
    }
//this->phi_analysis->dump();
    return false;
}

