//===---------SeparatePhiLiveRanges_Fer.cpp - Printing prototype ---------===//
//
//              Register Allocation via Coloring of Chordal Graphs
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it changes the CFG to avoid live ranges of phi-related virtuals
// been alive together.
//
// Date: February 21th, 2007
//
//===----------------------------------------------------------------------===//
//
// This pass separates the live ranges of phi-related virtuals. If the basic
// block contains a construct like:
// B3 := PHI(B1, B2);
// ...
// := B3, B1 // that is, B1 and B3 are alive past the phi-function.
// Then the method inserts a move instruction to make B1, and B3 unrelated:
// V := B1;
// B3 := PHI(V, B2);
//
//===----------------------------------------------------------------------===//

#include <iostream>                                      // cerr

#include "llvm/CodeGen/SSARegMap.h"

#include "llvm/CodeGen/MachineFunctionPass.h"            // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"

#include "llvm/fernando/NeatPrint_Fer.h"
#include "llvm/fernando/DefUseSites_Fer.h"
#include "llvm/fernando/EdgeLiveness_Fernando.h"
#include "llvm/fernando/SeparatePhiLiveRanges_Fer.h"       // class definition

using namespace llvm;

namespace {
    /// Currently, this pass is being required by the DefUseSites_Fer class,
    /// and by the EdgeLiveness_Fernando.h class.
    RegisterPass<SeparatePhiLiveRanges_Fer>
    X("separate_lr_phi", "Separate LR of phi-related virts - Fernando", true);
}


//===--------------------------------------------------------------------------
// This method creates a new virtual register with the same class as reg_model.
//===--------------------------------------------------------------------------
unsigned SeparatePhiLiveRanges_Fer::create_new_virtual_register
                                                         (unsigned reg_model) {
    // Take the register class.
    const TargetRegisterClass *trc =
                this->machine_function->getSSARegMap()->getRegClass(reg_model);
    SSARegMap * srm = this->machine_function->getSSARegMap();
    unsigned new_reg = srm->createVirtualRegister(trc);
    return new_reg;
}


//===--------------------------------------------------------------------------
// This method inserts a move instruction at the end of basic block mbb.
// The move instruction copies the value of register src to register dst.
// If the source of the move is a spill, then, instead of inserting a move
// instruction, this method directly loads the spilled value into the destiny.
//===--------------------------------------------------------------------------
MachineInstr * SeparatePhiLiveRanges_Fer::add_move(
    MachineBasicBlock & mbb,
    unsigned src,
    unsigned dst
) {
    assert(mbb.succ_size() == 1 && "Attempt to insert move in critical edge.");

    MachineBasicBlock::iterator iter = mbb.getFirstTerminator();

    const MRegisterInfo * reg_info =
                         this->machine_function->getTarget().getRegisterInfo();

    const TargetRegisterClass *rc =
                      this->machine_function->getSSARegMap()->getRegClass(src);

    reg_info->copyRegToReg(mbb, iter, dst, src, rc);
std::cerr << "Adding 2-phase move: " << dst << " := " << src << "\n";

    iter--;
    MachineInstr * new_inst = iter;
    return new_inst;
}


//===--------------------------------------------------------------------------
// This method returns true if the live ranges of 'live_in_reg' and 'v_reg'
// overlap inside the basic block 'mbb'.
//===--------------------------------------------------------------------------
bool SeparatePhiLiveRanges_Fer::overlap_in_mbb(
    unsigned live_in_reg,
    unsigned v_reg,
    MachineBasicBlock *mbb
) {
    // See if v_reg is used in mbb
    bool v_reg_is_used_in_mbb = false;
    std::vector< MachineInstr * > & v_use_sites =
                                     this->def_use_sites->get_use_sites(v_reg);
    for(unsigned u = 0; u < v_use_sites.size(); u++) {
        MachineInstr * mi = v_use_sites[u];        
        if(mi->getParent()->getNumber() == mbb->getNumber()) {
            if(mi->getOpcode() != TargetInstrInfo::PHI) {
                v_reg_is_used_in_mbb = true;
            }
        }
    }
    // If v_reg is not used inside mbb, we are done:
    if(!v_reg_is_used_in_mbb) {
        return false;
    }
    // Likewise, if live_in_reg is alive outside mbb, and v_reg is used inside
    // mbb, they interfere:
    if(this->edge_liveness->is_live_out(* mbb, live_in_reg)) {
        return true;
    }
    // Grab the uses of live_in_reg
    std::vector< MachineInstr * > & d_use_sites =
                               this->def_use_sites->get_use_sites(live_in_reg);
    bool lir_reg_is_used_in_mbb = false;
    for(unsigned u = 0; u < d_use_sites.size(); u++) {
        MachineInstr * mi = d_use_sites[u];        
        if(mi->getParent()->getNumber() == mbb->getNumber()) {
            assert(mi->getOpcode() != TargetInstrInfo::PHI &&
                 "Virtual def by phi-func is used by phi-func in the same BB");
            lir_reg_is_used_in_mbb = true;
        }
    }
    assert(lir_reg_is_used_in_mbb &&
                             "Virtual defined in phi-function is never used.");
    // Both virtuals are used in the basic block. First, find the last use of
    // live_in_reg:
    MachineBasicBlock::iterator last_use_iter = mbb->end();
    while (last_use_iter != mbb->begin()) {
        bool done = false;
        last_use_iter--;
        MachineInstr * mi = last_use_iter;
        for(unsigned t = 0; t < mi->getNumOperands(); t++) {
            MachineOperand & mo_aux = mi->getOperand(t);
            if(mo_aux.isRegister() && mo_aux.getReg() && mo_aux.isUse()) {
                if(mo_aux.getReg() == v_reg) {
                    done = true;
                    break;
                }
            }
        }
        if(done) {
            break;
        }
    }
    // See if there is any use of v_reg in the space between 'last_use_iter'
    // and mbb->begin(); If v_reg is defined in this space, then the live
    // ranges do not overlap.
    while (last_use_iter != mbb->begin()) {
        MachineInstr * mi = last_use_iter;
        bool done = false;
        for(unsigned t = 0; t < mi->getNumOperands(); t++) {
            MachineOperand & mo_aux = mi->getOperand(t);
            if(mo_aux.isRegister() && mo_aux.getReg()) {
                if(mo_aux.isUse()) {
                    if(mo_aux.getReg() == live_in_reg) {
                        done = true;
                    }
                } else if(mo_aux.isDef()) {
                    if(mo_aux.getReg() == v_reg) {
                        return false;
                    }
                }
            }
        }
        if(done) {
            return true;
        }
        last_use_iter--;
    }
    return false;
}


//===--------------------------------------------------------------------------
// To raise a parameter p of a phi-function means to replace it with a new
// virtual v, and, in order to keep the semantics of the old program, to insert
// a copy v := p in the basic block that feeds this parameter to the
// phi-function.
//===--------------------------------------------------------------------------
void SeparatePhiLiveRanges_Fer::raise_parameters(
    MachineInstr * phi,
    MachineBasicBlock * mbb
) {
    unsigned def_reg = phi->getOperand(0).getReg();
    bool def_live_out = this->edge_liveness->is_live_out(* mbb, def_reg);
    for(unsigned i = phi->getNumOperands() - 1; i >= 2; i -= 2) {
        unsigned use_reg = phi->getOperand(i-1).getReg();
        if( (this->edge_liveness->is_live_out(* mbb, use_reg) && def_live_out)
                             || this->overlap_in_mbb(def_reg, use_reg, mbb) ) {
            MachineOperand & mo = phi->getOperand(i-1);
            unsigned new_reg = create_new_virtual_register(use_reg);
            MachineBasicBlock *prd = phi->getOperand(i).getMachineBasicBlock();
            MachineInstr * move_inst = add_move(*prd, use_reg, new_reg);
            mo.setReg(new_reg);
            // Update liveness information.
//            this->edge_liveness->kill(*prd, *mbb, use_reg);
            this->edge_liveness->set_alive(*prd, *mbb, new_reg);
            // Update location:
            this->def_use_sites->set_def_site(move_inst, new_reg);
            this->def_use_sites->add_use_site(phi, new_reg);
            // Remove the use of the use_reg in phi
            this->def_use_sites->remove_use_site(phi, use_reg);
        }
    }
}


//===--------------------------------------------------------------------------
// reporst that this pass does not change anything on the control flow graph.
//===--------------------------------------------------------------------------
void SeparatePhiLiveRanges_Fer::getAnalysisUsage(AnalysisUsage & AU) const {
    AU.addRequired<DefUseSites_Fer>();
    AU.addRequired<EdgeLiveness_Fernando>();
    AU.setPreservesAll();
}


//===--------------------------------------------------------------------------
// runOnMachineFunction - Groups together the phi related virtuals.
//===--------------------------------------------------------------------------
bool SeparatePhiLiveRanges_Fer::runOnMachineFunction(MachineFunction & mf) {

    this->machine_function = & mf;

    // gets the liveness analysis information
    this->edge_liveness = & getAnalysis<EdgeLiveness_Fernando>();

    // gets the list of def/use sites
    this->def_use_sites = & getAnalysis<DefUseSites_Fer>();

    for (MachineFunction::iterator mbbi = mf.begin(), mbbe = mf.end();
                                                        mbbi != mbbe; ++mbbi) {
        MachineBasicBlock * mbb = mbbi;
        MachineBasicBlock::iterator mii = mbb->begin();
        std::vector<MachineInstr *> phi_insts;
        while (mii != mbb->end()) {
            MachineInstr * mi = mii;
            if(mi->getOpcode() == TargetInstrInfo::PHI) {
                this->raise_parameters(mi, mbb);
            } else {
                // Always safe, because phi functions are always the first
                // instructions in the basic block.
                break;
            }
            mii++;
        }
    }
//this->def_use_sites->dump();
    return false;
}
