//===------------ SplitPhiGroups_Fer.cpp - Printing prototype -------------===//
//
//              Register Allocation via Coloring of Chordal Graphs
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it split live ranges of phi-related virtual registers.
//
// Date: February 18th, 2007
//
//===----------------------------------------------------------------------===//
//
// This pass splits the live ranges of virtual registers that contribute
// parameters to phi-functions in the same column of a phi-matrix. For example,
// if the target code contains an assignment such as:
// A1 = PHI(V, A3);
// A2 = PHI(V, A4);
// The the method will split the live range of V in two interfering live ranges
// by inserting two move instructions at the end of the basic block that feeds
// the first column of the matrix. After the method, the program will be like:
// A5 = V;
// A6 = V;
// goto ...
// A1 = PHI(A5, A3);
// A2 = PHI(A6, A4);
//
//===----------------------------------------------------------------------===//

#include <iostream>                                      // cerr

#include "llvm/CodeGen/SSARegMap.h"

#include "llvm/CodeGen/MachineFunctionPass.h"            // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"

#include "llvm/fernando/NeatPrint_Fer.h"
#include "llvm/fernando/CriticalEdgeRemoval_Fer.h"
#include "llvm/fernando/SplitPhiGroups_Fer.h"       // class definition

using namespace llvm;

namespace {
    /// Currently, this pass is being required by the DefUseSites_Fer class,
    /// and by the EdgeLiveness_Fernando.h class.
    RegisterPass<SplitPhiGroups_Fer>
    X("split_phi", "Split Phi Groups - Fernando", true);
}


//===--------------------------------------------------------------------------
// This method creates a new virtual register with the same class as reg_model.
//===--------------------------------------------------------------------------
unsigned SplitPhiGroups_Fer::create_new_virtual_register(unsigned reg_model) {
    // Take the register class.
    const TargetRegisterClass *trc =
                this->machine_function->getSSARegMap()->getRegClass(reg_model);
    SSARegMap * srm = this->machine_function->getSSARegMap();
    unsigned new_reg = srm->createVirtualRegister(trc);
    this->num_v_regs++;
    return new_reg;
}


//===--------------------------------------------------------------------------
// This method inserts a move instruction at the end of basic block mbb.
// The move instruction copies the value of register src to register dst.
// If the source of the move is a spill, then, instead of inserting a move
// instruction, this method directly loads the spilled value into the destiny.
//===--------------------------------------------------------------------------
void SplitPhiGroups_Fer::add_move(
    MachineBasicBlock & mbb,
    unsigned src,
    unsigned dst
) {
    assert(mbb.succ_size() == 1 && "Attempt to insert move in critical edge.");

    MachineBasicBlock::iterator iter = mbb.getFirstTerminator();

    const MRegisterInfo * reg_info =
                         this->machine_function->getTarget().getRegisterInfo();

    const TargetRegisterClass *rc =
                      this->machine_function->getSSARegMap()->getRegClass(src);

    reg_info->copyRegToReg(mbb, iter, dst, src, rc);

    iter--;
    MachineInstr * new_instr = iter;
    this->split_points.push_back(new_instr);

//std::cerr << "Adding move: " << dst << " := " << src << "\n";
}


//===--------------------------------------------------------------------------
// reporst that this pass does not change anything on the control flow graph.
//===--------------------------------------------------------------------------
void SplitPhiGroups_Fer::getAnalysisUsage(AnalysisUsage & AU) const {
    AU.addRequired<CriticalEdgeRemoval_Fer>();
    AU.setPreservesAll();
}


//===--------------------------------------------------------------------------
// This function takes a phi-function and the basic block where the phi
// function is defined, and adds one copy instruction for each parameter of the
// phi-function.
//===--------------------------------------------------------------------------
void SplitPhiGroups_Fer::raise_parameters(
    MachineInstr * phi,
    MachineBasicBlock *mbb
) {
    for ( int i = phi->getNumOperands() - 1; i >= 2; i -= 2 ) {
        MachineBasicBlock * pred = phi->getOperand(i).getMachineBasicBlock();
        MachineOperand & mo = phi->getOperand(i-1);
        unsigned use_reg = mo.getReg();
        unsigned new_reg = create_new_virtual_register(use_reg);
        add_move(*pred, use_reg, new_reg);
        mo.setReg(new_reg);
    }
}


//===--------------------------------------------------------------------------
// This method prints all the instrs. stored in the vector of split_points.
//===--------------------------------------------------------------------------
void SplitPhiGroups_Fer::dump() {
    std::cerr << "Split points:\n";
    for(unsigned u = 0; u < this->split_points.size(); u++) {
        MachineInstr * mi = this->split_points[u];
        std::cerr << NeatPrint_Fer::mi2string(*mi) << "\n";
    }
}


//===--------------------------------------------------------------------------
// runOnMachineFunction - Groups together the phi related virtuals.
//===--------------------------------------------------------------------------
bool SplitPhiGroups_Fer::runOnMachineFunction(MachineFunction & mf) {

    this->machine_function = & mf;

    SSARegMap * srm = mf.getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    this->num_v_regs = last_v_reg - MRegisterInfo::FirstVirtualRegister;

    for (MachineFunction::iterator mbbi = mf.begin(), mbbe = mf.end();
                                                        mbbi != mbbe; ++mbbi) {
        MachineBasicBlock * mbb = mbbi;
        MachineBasicBlock::iterator mii = mbb->begin();
        while (mii != mbb->end()) {
            MachineInstr * mi = mii;
            if(mi->getOpcode() == TargetInstrInfo::PHI) {
                this->raise_parameters(mi, mbb);
            } else {
                // Always safe, because phi functions are always the first
                // instructions in the basic block.
                break;
            }
            mii++;
        }
    }
//this->dump();
    return false;
}
