//===----------- ThreeAddr2TwoAddr_Fer.cpp - Printing prototype -----------===//
//
//             Register Allocation via Coloring of Chordal Graphs 
//
// This pass converts a three address instruction to a two address instruction.
//
// Date: May 28th, 2006
//
//===----------------------------------------------------------------------===//
//
// In LLVM, all the instructions are three address code, but some target
// architectures have two address instructions. Therefore, it is necessary to
// make this conversion, what is exaclty what this class does.
//
//===----------------------------------------------------------------------===//

#include <iostream>                              // cerr

#include "llvm/CodeGen/Passes.h"                 // EdgeLivenessFernandoPassID
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/Instructions.h"                   // MachineOperand
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include <llvm/Target/TargetMachine.h>

#include "llvm/fernando/NeatPrint_Fer.h" // class definition
#include "llvm/fernando/ThreeAddr2TwoAddr_Fer.h" // class definition

using namespace llvm;


namespace {
    RegisterPass<ThreeAddr2TwoAddr_Fer>
    X("addr3_addr2", "3 address to 2 address convertion - Fernando.");
}


//===--------------------------------------------------------------------------
// Finds the two address instructions in the basic block and converts them to
// two address instructions. We assume that, if the instruction is in the
// format a := b + a, then it is commutative. We also assume that the first
// three operands of the instruction are registers.
//===--------------------------------------------------------------------------
bool ThreeAddr2TwoAddr_Fer::remove3address(MachineBasicBlock & mbb) {
    bool made_change = false;
    for (MachineBasicBlock::iterator mii = mbb.begin(), mei = mbb.end();
                                                           mii != mei; ++mii) {
        // Get the instruction data:
        MachineInstr * mi = mii;
        unsigned opcode = mi->getOpcode();
        const TargetInstrDescriptor & tid = this->tii->get(opcode);

        if(this->tii->isTwoAddrInstr(opcode)) {
            if(tid.Flags & M_CONVERTIBLE_TO_3_ADDR) {
                // It is possible to convert the instruction to 3-address code.
                // So, go ahead and do it.
                MachineInstr * new_inst = tii->convertToThreeAddress(mi);
                if (new_inst) {
                    mbb.insert(mii, new_inst);
                    mbb.erase(mii);
                    assert(!tii->isTwoAddrInstr(new_inst->getOpcode()) &&
                             "convertToThreeAddress returned a 2-addr instr?");
                    continue;
                }
            }
            // Ok, the instruction will have to be changed manually:
            convert_to_2_addr_plus_copy(mii, mbb);
            made_change = true;
        }
    }
    return made_change;
}


//===--------------------------------------------------------------------------
// This method takes an instruction in three address code format, and converts
// it two two instructions: a copy, and a second instruction in two address
// code format. Ex.: a := b + c => a := b; a += c;
//===--------------------------------------------------------------------------
void ThreeAddr2TwoAddr_Fer::convert_to_2_addr_plus_copy(
    MachineBasicBlock::iterator mii,
    MachineBasicBlock & mbb
) {

    MachineInstr * mi = mii;

    // Start by checking that this instruction is in the expected
    // format a := b + c
    assert(mi->getOperand(0).isRegister() && mi->getOperand(0).getReg() &&
               mi->getOperand(0).isDef() && "Invalid two address instruction");
    assert(mi->getOperand(1).isRegister() && mi->getOperand(1).getReg() &&
               mi->getOperand(1).isUse() && "Invalid two address instruction");

    // This pass only checks virtual registers:
    assert(MRegisterInfo::isVirtualRegister(mi->getOperand(0).getReg()) &&
        "Only instructions with virtual operands can be lowered into 3-addr.");
    assert(MRegisterInfo::isVirtualRegister(mi->getOperand(1).getReg()) &&
        "Only instructions with virtual operands can be lowered into 3-addr.");

    unsigned virt_a = mi->getOperand(0).getReg();
    unsigned virt_b = mi->getOperand(1).getReg();

    // a := b + c -> a := b, a := a + c
    const TargetRegisterClass* rc =
                                 this->mf->getSSARegMap()->getRegClass(virt_a);
    // Inserting a := b
    this->ri->copyRegToReg(mbb, mii, virt_a, virt_b, rc);
    // replace all occurences of b with a
    for (unsigned i = 1, e = mi->getNumOperands(); i != e; i++) {
        if (mi->getOperand(i).isRegister() &&
                                        mi->getOperand(i).getReg() == virt_b) {
            mi->getOperand(i).setReg(virt_a);
        }
    }

    // After the previous transformations, we have a := a + x. We can
    // remove the second operator, and change the first into a use.
    assert(mi->getOperand(0).isDef() && mi->getOperand(1).isUse()
                                                && "Wrong instruction format");
    mi->getOperand(1).setReg(mi->getOperand(0).getReg());
}


//===--------------------------------------------------------------------------
// Checks each instruction in the function, and converts the three address
// instructions to two address if necessary.
//===--------------------------------------------------------------------------
bool ThreeAddr2TwoAddr_Fer::runOnMachineFunction(MachineFunction & mf) {
    this->mf = & mf;
    const TargetMachine & tm = mf.getTarget();
    this->ri = tm.getRegisterInfo();
    this->tii = tm.getInstrInfo();
    bool hasChanged = false;

    // traverse the control flow graph in depth first order pinpointing the
    // registers alive in each edge:
    for (MachineFunction::iterator mbb = mf.begin(), mbbe = mf.end();
                                                          mbb != mbbe; ++mbb) {
        bool result = remove3address(*mbb);
        hasChanged = result || hasChanged;
    }

    return hasChanged;
}
