//===------------- Virtual_Fernando.cpp - Virtual Register ----------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it represents a virtual register.
//
// Date: January 29th, 2007
//
//===----------------------------------------------------------------------===//
//
// This file contains information referent to a virtual register. Such
// information is necessary during register allocation, and includes, for
// instance, the size of the live range of that virtual, the instruction where
// the virtual is defined, the instructions where it is used, etc.
//
//===----------------------------------------------------------------------===//

#include <vector>
#include <sstream>                               // ostringstream (RegClique)

#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/CodeGen/MachineFunctionPass.h"     // MachineBasicBlock


#include "llvm/fernando/Virtual_Fernando.h"

using namespace llvm;

std::string Virtual_Fernando::to_string() const {
    std::ostringstream aux;
    aux << "to string method.";
    return aux.str();
}


//===--------------------------------------------------------------------------
// Returns the instruction where the temporary is defined. This instruction is
// unique, because the target program is expected to be in the SSA-form.
//===--------------------------------------------------------------------------
const MachineInstr * Virtual_Fernando::get_def_site() const {
    assert(this->def != NULL &&
               "Operand has no definition site. Maybe it is a global variable");
    return this->def;
}


//===--------------------------------------------------------------------------
// This method removes the instruction mi from the list of use sites of this
// virtual.
//===--------------------------------------------------------------------------
void Virtual_Fernando::remove_use_site(MachineInstr * mi) {

    bool found = false;

    for(unsigned u = 0; u < this->uses.size(); u++) {
        if(this->uses[u] == mi) {
            MachineInstr * aux = this->uses[u];
            this->uses[u] = this->uses[this->uses.size()-1];
            this->uses[this->uses.size()-1] = aux;
            found = true;
            break;
        }
    }

    if(found) {
        this->uses.pop_back();
    }
}


//===--------------------------------------------------------------------------
// Computes the e-th power of n, that is, n**e;
//===--------------------------------------------------------------------------
unsigned pow(unsigned n, unsigned e) {
    unsigned p = 1;
    for(unsigned i = 0; i < e; i++) {
        p = p * n;
    }
    return p;
}


//===--------------------------------------------------------------------------
// Returns the spilling weight of the given temporary. Virtual registers with
// smaller weights are better candidates to be spilled. The heuristic function
// is computed according to the following algorithm:
//
// Instruction def_i := definition_site(v);
//
// Unsigned w := 1 + 10 ** loop_depth[def_i];
// 
// foreach ( Instruction i in use_sites(v) )
// |
// |    w := w + 1 + 10 ** loop_depth[i];
// |
// end foreach
//
// v2w(v) := w / live_range_size;
// 
//===--------------------------------------------------------------------------
const void Virtual_Fernando::compute_weight(LoopInfo & li) {
    const MachineBasicBlock * mbb_def = this->def->getParent();
    const BasicBlock * bb = mbb_def->getBasicBlock();
    unsigned nesting_factor = li.getLoopDepth(bb);
    unsigned w = pow(10, nesting_factor);
    for(unsigned i = 0; i < this->uses.size(); i++) {
        MachineInstr * use_site = this->uses[i];
        MachineBasicBlock * mbb_use = use_site->getParent();
        bb = mbb_use->getBasicBlock();
        nesting_factor = li.getLoopDepth(bb);
        w = w + pow(10, nesting_factor);
    }
    this->weight = (w / this->live_range_size);
}
